import shutil
import typer
import ast
import subprocess
import os


def get_function_refs(file_path):
    load_ref, infer_ref, unload_ref = infer_function_refs(file_path)
    if load_ref is not None:
        load_ref = typer.prompt("Enter the load function reference", default=load_ref)
    else:
        load_ref = typer.prompt("Enter the load function reference [Ex: TritonClass.load]")

    if infer_ref is not None:
        infer_ref = typer.prompt("Enter the infer function reference", default=infer_ref)
    else:
        infer_ref = typer.prompt("Enter the infer function reference [Ex: TritonClass.infer]")

    if unload_ref is not None:
        unload_ref = typer.prompt("Enter the unload function reference", default=unload_ref)
    else:
        unload_ref = typer.prompt("Enter the unload function reference [Ex: TritonClass.unload]", default="None")

    return load_ref, infer_ref, unload_ref


def infer_function_refs(file_path):
    with open(file_path, "r") as f:
        data = f.read()
    tree = ast.parse(data)
    load_ref = None
    infer_ref = None
    unload_ref = None
    for node in ast.walk(tree):
        for child in ast.iter_child_nodes(node):
            child.parent = node
    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef):
            if node.name == "load":
                load_ref = f"{node.parent.name}.{node.name}"
            elif node.name == "infer":
                infer_ref = f"{node.parent.name}.{node.name}"
            elif node.name == "unload":
                unload_ref = f"{node.parent.name}.{node.name}"
    return load_ref, infer_ref, unload_ref


def build_image(project_path, tag, triton_version, requirements_file):
    try:
        if requirements_file != "None":
            shutil.copy(requirements_file, f"{project_path}/1/requirements.txt")
        file = open(f"{project_path}/Dockerfile", "r")
        lines = file.readlines()

        for i in range(len(lines)):
            lines[i] = lines[i].replace("##tritonversion##", triton_version)
            lines[i] = lines[i].replace('##model_name##', project_path.split('/')[-1])

        file = open(f"{project_path}/Dockerfile", "w")
        file.writelines(lines)
        file.close()

        # check requirement.txt file exists
        if not os.path.exists(f"{project_path}/1/requirements.txt"):
            # create an empty requirements.txt file
            with open(f"{project_path}/1/requirements.txt", "w") as f:
                f.write("")

        docker_build_cmd = "docker build -t {tag} {dir} -f {dir}/Dockerfile".format(
            dir=project_path, tag=tag
        )
        completed_process = subprocess.run(
            docker_build_cmd.split(" "), stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True
        )
    except subprocess.CalledProcessError as e:
        raise RuntimeError("command '{}' return with error (code {}): {}, error: {}".format(e.cmd, e.returncode, e.output, e.stderr))
    except Exception as e:
        raise SystemExit(f"Error: {e}")



import os
import sys
import ##file## as app
import json
import triton_python_backend_utils as pb_utils
import numpy as np
import uuid

custom_model = app.##model_class##()

model_data_type_maping = {
    bool: "TYPE_BOOL",
    np.uint8: "TYPE_UINT8",
    np.uint16: "TYPE_UINT16",
    np.uint32: "TYPE_UINT32",
    np.uint64: "TYPE_UINT64",
    np.int8: "TYPE_INT8",
    np.int16: "TYPE_INT16",
    np.int32: "TYPE_INT32",
    np.int64: "TYPE_INT64",
    np.float16: "TYPE_FP16",
    np.float32: "TYPE_FP32",
    np.float64: "TYPE_FP64",
    str: "TYPE_STRING",
    float: "TYPE_FP64",
    int: "TYPE_INT64",
}


model_triton_data_type_maping = {
    bool: "BOOL",
    np.uint8: "UINT8",
    np.uint16: "UINT16",
    np.uint32: "UINT32",
    np.uint64: "UINT64",
    np.int8: "INT8",
    np.int16: "INT16",
    np.int32: "INT32",
    np.int64: "INT64",
    np.float16: "FP16",
    np.float32: "FP32",
    np.float64: "FP64",
    str: "BYTES",
    float: "FP64",
    int: "INT64",
}

try:
    import tensorflow as tf

    model_data_type_maping[tf.bfloat16] = "TYPE_BF16"
    model_triton_data_type_maping[tf.bfloat16] = "BF16"
except ImportError:
    # Code to handle the absence of TensorFlow here
    pass


is_validator = os.environ.get("IS_VALIDATOR")


class ConfigCreationException(Exception):
    def __init__(self, message):
        self.message = message


class TritonPythonModel:
    def initialize(self, args):
        custom_model.##load_func##()

    def execute(self, requests):
        responses = []
        for request in requests:
            required_input = ["##input_list##"]
            inputs = {}
            for each in required_input:
                input_numpy_tensor = pb_utils.get_input_tensor_by_name(
                    request, each["name"]
                )
                if (
                    "optional" in each
                    and each["optional"] == True
                    and input_numpy_tensor is None
                ):
                    continue

                input_numpy_tensor = input_numpy_tensor.as_numpy()
                size = len(input_numpy_tensor)
                shape = input_numpy_tensor.shape
                if shape == (1,):
                    if each["datatype"] == "BYTES":
                        inputs[each["name"]] = np.vectorize(lambda x: x.decode('utf-8'))(input_numpy_tensor)[0]
                    else:
                        inputs[each["name"]] = input_numpy_tensor[0]
                else:
                    input_array = []
                    for i in range(0, size):
                        if each["datatype"] == "BYTES":
                            input_array.append(
                                np.vectorize(lambda x: x.decode('utf-8'))(input_numpy_tensor[i])
                            )
                        else:
                            input_array.append(input_numpy_tensor[i])
                    inputs[each["name"]] = input_array

            inference_outputs = custom_model.##infer_func##(inputs)
            required_output = ["#output_list#"]
            if len(required_output) == 0:
                required_output = []
                try:
                    required_output = TritonPythonModel.create_config_from_json(
                        self, inference_outputs
                    )
                except ValueError as e:
                    raise ConfigCreationException(str(e))
                except Exception as e:
                    raise ConfigCreationException(
                        "Exception occurred while creating output config " + str(e)
                    )
            outputs = []
            for each in required_output:
                if each["name"] in inference_outputs:
                    each_inference_output = inference_outputs[each["name"]]
                    if isinstance(each_inference_output, list):
                        temp_inference_output = []
                        for each_element in each_inference_output:
                            if isinstance(each_element, dict) or isinstance(
                                each_element, tuple
                            ):
                                raise ValueError(
                                    "Output "
                                    + each["name"]
                                    + " is of type dictionary or tuple which is currently not supported. You can try json.dumps("
                                    + each["name"]
                                    + ")"
                                )
                            if each["datatype"] == "BYTES":
                                temp_inference_output.append(
                                    each_element.encode("utf8")
                                )
                            else:
                                temp_inference_output.append(each_element)
                        np_inference_output = np.array(temp_inference_output)
                        outputs.append(
                            pb_utils.Tensor(each["name"], np_inference_output)
                        )
                    elif isinstance(each_inference_output, dict) or isinstance(
                        each_inference_output, tuple
                    ):
                        raise ValueError(
                            "Output "
                            + each["name"]
                            + " is of type dictionary or tuple which is currently not supported. You can try json.dumps("
                            + each["name"]
                            + ")"
                        )
                    else:
                        temp_inference_output = None
                        if each["datatype"] == "BYTES":
                            temp_inference_output = np.array(
                                [each_inference_output.encode("utf8")]
                            )
                        else:
                            temp_inference_output = np.array([each_inference_output])
                        outputs.append(
                            pb_utils.Tensor(each["name"], temp_inference_output)
                        )
                else:
                    raise ValueError(
                        "Requried Output "
                        + each["name"]
                        + " not found in the inference response"
                    )
            inference_response = pb_utils.InferenceResponse(output_tensors=outputs)
            responses.append(inference_response)
        return responses

    def finalize(self, args):
        custom_model.##unload_func##(args)

    def create_config_from_json(self, output):
        current_directory = os.path.dirname(os.path.abspath(__file__))
        parent_directory = os.path.dirname(current_directory)
        output_config = []
        output_config_for_return = []
        for key in output:
            each_output = {}
            each_output_for_return = {}
            each_output["name"] = "#" + key + "#"
            each_output_for_return["name"] = key

            type_of_output = TritonPythonModel.get_type(self, output[key])
            if type_of_output == dict or type_of_output == tuple:
                raise ValueError(
                    "Output "
                    + key
                    + " is of type dictionary or tuple which is currently not supported. You can try json.dumps("
                    + key
                    + ")"
                )
            if type_of_output == list:
                raise ValueError(
                    "Output "
                    + key
                    + " is empty list which cannot be used to infer the data type. Please provide a non-empty list as output."
                )
            each_output["data_type"] = TritonPythonModel.map_data_type(
                self, type_of_output
            )
            each_output_for_return["datatype"] = TritonPythonModel.map_triton_data_type(
                self, type_of_output
            )

            shape_of_output = TritonPythonModel.get_shape(self, output[key])
            if len(shape_of_output) > 0:
                each_output["dims"] = shape_of_output
                each_output_for_return["dims"] = shape_of_output
            else:
                each_output["dims"] = [1]
                each_output_for_return["dims"] = [1]

            output_config.append(each_output)
            output_config_for_return.append(each_output_for_return)

        if is_validator == "True":
            config_path = os.path.join(parent_directory, "config.pbtxt")

            with open(config_path, "r") as file:
                content = file.read()

            modified_content = content.replace(
                "output []",
                "output "
                + json.dumps(output_config).replace('"', "").replace("#", '"'),
            )

            with open(config_path, "w", encoding="utf-8") as file:
                file.write(modified_content)

        return output_config_for_return

    def map_data_type(self, variable_type):
        return model_data_type_maping[variable_type]

    def map_triton_data_type(self, variable_type):
        return model_triton_data_type_maping[variable_type]

    def get_shape(self, value):
        shape = []

        if isinstance(value, list):
            shape.append(len(value))
            if shape[0] > 0 and isinstance(value[0], list):
                shape.extend(TritonPythonModel.get_shape(self, value[0]))
        elif isinstance(value, np.ndarray):
            shape.extend(value.shape)

        return shape

    def get_type(self, variable):
        if not isinstance(variable, (list, np.ndarray)):
            return type(variable)
        elif isinstance(variable, (list, np.ndarray)) and len(variable) > 0:
            return TritonPythonModel.get_type(self, variable[0])
        elif isinstance(variable, (list, np.ndarray)) and len(variable) == 0:
            return type([])
        elif not variable:
            return type(None)
        else:
            return TritonPythonModel.get_type(self, variable[0])



system_prompt_config_pbtxt = """
                Your task is to generate config.pbtxt for the given code snippet in correct format for nvidia triton server.
                User may also provide input.json, output.json and input_schema.py. Infer the necessary information from these files.
                Below are two examples for the code snippet
                ## Example 1
                from threading import Thread
                from typing import Iterator
                import os

                import torch
                import json
                from transformers import AutoModelForCausalLM, AutoTokenizer, TextIteratorStreamer

                model_id = "meta-llama/Llama-2-7b-chat-hf"


                class PythonModel:
                    def get_prompt(self, message, chat_history, system_prompt):
                        texts = [f"[INST] <<SYS>>\n{system_prompt}\n<</SYS>>\n\n"]
                        chat_history = json.loads(chat_history)

                        for each in chat_history:
                            print("each", each, flush=True)
                            user_input = each["user_input"]
                            response = each["response"]
                            texts.append(
                                f"{user_input.strip()} [/INST] {response.strip()} </s><s> [INST] "
                            )
                        texts.append(f"{message.strip()} [/INST]")
                        return "".join(texts)

                    def get_input_token_length(self, message, chat_history, system_prompt):
                        prompt = self.get_prompt(message, chat_history, system_prompt)
                        input_ids = self.tokenizer([prompt], return_tensors="np")["input_ids"]
                        return input_ids.shape[-1]

                    def run_function(
                        self,
                        message,
                        chat_history,
                        system_prompt,
                        max_new_tokens=1024,
                        temperature=0.8,
                        top_p=0.95,
                        top_k=5,
                    ):
                        prompt = self.get_prompt(message, chat_history, system_prompt)
                        inputs = self.tokenizer([prompt], return_tensors="pt").to("cuda")

                        streamer = TextIteratorStreamer(
                            self.tokenizer, timeout=10.0, skip_prompt=True, skip_special_tokens=True
                        )
                        generate_kwargs = dict(
                            inputs,
                            streamer=streamer,
                            max_new_tokens=max_new_tokens,
                            do_sample=True,
                            top_p=top_p,
                            top_k=top_k,
                            temperature=temperature,
                            num_beams=1,
                        )
                        t = Thread(target=self.model.generate, kwargs=generate_kwargs)
                        t.start()

                        outputs = ""
                        for text in streamer:
                            outputs += text

                        return outputs

                    def load(self):
                        token = "abscafs"
                        self.tokenizer = AutoTokenizer.from_pretrained(
                            model_id, use_auth_token=token, device="cuda"
                        )
                        if torch.cuda.is_available():
                            self.model = AutoModelForCausalLM.from_pretrained(
                                model_id,
                                torch_dtype=torch.float16,
                                device_map="auto",
                                use_auth_token=token,
                            )
                        else:
                            self.model = None

                    def infer(self, inputs):
                        message = inputs["message"]
                        chat_history = inputs["chat_history"] if "chat_history" in inputs else []
                        system_prompt = inputs["system_prompt"] if "system_prompt" in inputs else ""
                        result = self.run_function(
                            message=message,
                            chat_history=chat_history,
                            system_prompt=system_prompt,
                        )
                        return {"generated_text": result}

                    def unload(self):
                        self.tokenizer = None
                        self.model = None


                config.pbtxt

                name: "new_llama2_7b_chat_0b2812185e684a56b7a9f15b68cbe6ee"
                backend: "python"
                input [
                    {
                        name: "message", 
                        dims: [1], 
                        data_type: TYPE_STRING
                    }, 
                    {
                        name: "chat_history", 
                        dims: [1], 
                        data_type: TYPE_STRING
                    }, 
                    {
                        name: "system_prompt", 
                        dims: [1], 
                        data_type: TYPE_STRING
                    }
                ]
                output [
                    {
                        name: "generated_text", 
                        dims: [1], 
                        data_type: TYPE_STRING
                    }
                ]
                instance_group [
                    {
                        kind: KIND_GPU
                    }
                ]


                ## Example 2
                import os

                os.environ[
                    "TRANSFORMERS_CACHE"
                ] = "/opt/tritonserver/model_repository/falcon7b/hf_cache"
                import json

                import numpy as np
                import torch
                import transformers
                import triton_python_backend_utils as pb_utils


                class TritonPythonModel:
                    def initialize(self, args):
                        self.logger = pb_utils.Logger
                        self.model_config = json.loads(args["model_config"])
                        self.model_params = self.model_config.get("parameters", {})
                        default_hf_model = "tiiuae/falcon-7b"
                        default_max_gen_length = "15"
                        # Check for user-specified model name in model config parameters
                        hf_model = self.model_params.get("huggingface_model", {}).get(
                            "string_value", default_hf_model
                        )
                        # Check for user-specified max length in model config parameters
                        self.max_output_length = int(
                            self.model_params.get("max_output_length", {}).get(
                                "string_value", default_max_gen_length
                            )
                        )

                        self.logger.log_info(f"Max sequence length: {self.max_output_length}")
                        self.logger.log_info(f"Loading HuggingFace model: {hf_model}...")
                        # Assume tokenizer available for same model
                        self.tokenizer = transformers.AutoTokenizer.from_pretrained(hf_model)
                        self.pipeline = transformers.pipeline(
                            "text-generation",
                            model=hf_model,
                            torch_dtype=torch.float16,
                            tokenizer=self.tokenizer,
                            device_map="auto",
                        )
                        self.pipeline.tokenizer.pad_token_id = self.tokenizer.eos_token_id

                    def execute(self, requests):
                        prompts = []
                        for request in requests:
                            input_tensor = pb_utils.get_input_tensor_by_name(request, "text_input")
                            multi_dim = input_tensor.as_numpy().ndim > 1
                            if not multi_dim:
                                prompt = input_tensor.as_numpy()[0].decode("utf-8")
                                self.logger.log_info(f"Generating sequences for text_input: {prompt}")
                                prompts.append(prompt)
                            else:
                                # Implementation to accept dynamically batched inputs
                                num_prompts = input_tensor.as_numpy().shape[0]
                                for prompt_index in range(0, num_prompts):
                                    prompt = input_tensor.as_numpy()[prompt_index][0].decode("utf-8")
                                    prompts.append(prompt)

                        batch_size = len(prompts)
                        return self.generate(prompts, batch_size)

                    def generate(self, prompts, batch_size):
                        sequences = self.pipeline(
                            prompts,
                            max_length=self.max_output_length,
                            pad_token_id=self.tokenizer.eos_token_id,
                            batch_size=batch_size,
                        )
                        responses = []
                        texts = []
                        for i, seq in enumerate(sequences):
                            output_tensors = []
                            text = seq[0]["generated_text"]
                            texts.append(text)
                            tensor = pb_utils.Tensor("text_output", np.array(texts, dtype=np.object_))
                            output_tensors.append(tensor)
                            responses.append(pb_utils.InferenceResponse(output_tensors=output_tensors))

                        return responses

                    def finalize(self):
                        print("Cleaning up...")

                config.pbtxt
                
                name: "falcon-7basdfvdawreq7a9f15b68cbe6ee"
                backend: "python"
                parameters: {
                    key: "huggingface_model",
                    value: {string_value: "tiiuae/falcon-7b"}
                }
                parameters: {
                    key: "max_output_length",
                    value: {string_value: "15"}
                }
                input [
                    {
                        name: "text_input"
                        data_type: TYPE_STRING
                        dims: [ 1 ]
                    }
                ]
                output [
                    {
                        name: "text_output"
                        data_type: TYPE_STRING
                        dims: [ -1 ]
                    }
                ]
                instance_group [
                    {
                        kind: KIND_GPU
                    }
                ]
            
            In your response, Just provide the code snippet without any quotes or comments
            """



import os

HOME = os.path.expanduser("~")
CONFIG_FILE_PATH = os.path.join(HOME, ".triton_copilot/config")


def is_initialized():
    return os.path.exists(CONFIG_FILE_PATH)


def create_config_file():
    directory_path = os.path.join(HOME, ".triton_copilot")
    os.makedirs(directory_path)
    #  create an empty config file
    with open(CONFIG_FILE_PATH, "w") as f:
        f.write("")


def save_openai_tokens(openai_key, openai_org):
    # if config file does not exist, create it
    if not is_initialized():
        create_config_file()
    #  save the tokens to the config file
    with open(CONFIG_FILE_PATH, "w") as f:
        f.write(f"OPENAI_KEY={openai_key}\n")
        if openai_org:
            f.write(f"OPENAI_ORG={openai_org}\n")


def save_claude_tokens(claude_api_key):
    # if config file does not exist, create it
    if not is_initialized():
        create_config_file()
    #  save the tokens to the config file
    with open(CONFIG_FILE_PATH, "w") as f:
        f.write(f"CLAUDE_API_KEY={claude_api_key}\n")


def get_model():
    openai_key, _ = get_openai_keys()
    claude_api_key = get_claude_keys()
    if openai_key:
        return "gpt4"
    elif claude_api_key:
        return "claude3"


def get_claude_keys():
    api_key = None
    with open(CONFIG_FILE_PATH, "r") as f:
        lines = f.readlines()
        for line in lines:
            if "CLAUDE_API_KEY" in line:
                api_key = line.split("=")[1].strip()

    return api_key


def get_openai_keys():
    api_key = None
    org_id = None
    with open(CONFIG_FILE_PATH, "r") as f:
        lines = f.readlines()
        for line in lines:
            if "OPENAI_KEY" in line:
                api_key = line.split("=")[1].strip()
            elif "OPENAI_ORG" in line:
                org_id = line.split("=")[1].strip()

    return api_key, org_id





import ast
import subprocess
import typer


def validate_refs(load_ref, infer_ref, unload_ref, file_path):
    try:
        if load_ref == "None" or len(load_ref.split(".")) != 2:
            typer.secho("Error: load function reference is missing or incorrect format", fg=typer.colors.BRIGHT_RED)
            raise typer.Exit()
        if infer_ref == "None" or len(infer_ref.split(".")) != 2:
            typer.secho("Error: infer function reference is missing or incorrect format", fg=typer.colors.BRIGHT_RED)
            raise typer.Exit()

        if len(unload_ref.split(".")) != 2:
            if unload_ref == "None":
                unload_ref = None
            else:
                typer.secho("Error: unload function reference is incorrect format", fg=typer.colors.BRIGHT_RED)
                raise typer.Exit()

        #  check if the references are present in the file
        with open(file_path, "r") as f:
            data = f.read()
        tree = ast.parse(data)
        for ref in [load_ref, infer_ref, unload_ref]:
            if ref is not None:
                class_name = ref.split(".")[0]
                func_name = ref.split(".")[1]
                found = False
                for node in ast.walk(tree):
                    if isinstance(node, ast.ClassDef):
                        if node.name == class_name:
                            for child in ast.walk(node):
                                if isinstance(child, ast.FunctionDef):
                                    if child.name == func_name:
                                        found = True
                                        break
                    if found:
                        break
                if not found:
                    raise SystemExit(f"Error: {ref} not found in the file")

        # check if all three references are present in same class or not
        if infer_ref.split(".")[0] != load_ref.split(".")[0]:
            raise SystemExit("Error: infer function should be in the same class as load function")
        if unload_ref is not None and unload_ref.split(".")[0] != load_ref.split(".")[0]:
            raise SystemExit("Error: unload function should be in the same class as load function")
    except Exception as e:
        typer.secho(f"Failed to validate references: {e}", fg=typer.colors.BRIGHT_RED)
        raise typer.Exit()


def check_gpu():
    try:
        result = subprocess.run(['nvidia-smi'], stdout=subprocess.PIPE)
        if result.returncode == 0:
            typer.echo("GPU is available!")
            return None
        else:
            return "No GPU available, using CPU instead."
    except FileNotFoundError:
        return "nvidia-smi command not found. Please ensure NVIDIA drivers are installed."


def check_docker():
    try:
        result = subprocess.run(['docker', 'info'], stdout=subprocess.PIPE)
        if result.returncode == 0:
            typer.echo("Docker is installed!")
            return None
        else:
            return "Docker is not installed."
    except FileNotFoundError:
        return "Docker command not found. Please install Docker."


def check_dependencies():
    msg = ""
    # check if GPU is available
    err = check_gpu()
    if err:
        msg = err
    # check if docker is installed
    err = check_docker()
    if err:
        msg += "\n" + err

    if msg != "":
        return msg
    return None


def validate_file(file_path):
    try:
        with open(file_path, "r") as f:
            data = f.read()
        ast.parse(data)
    except IsADirectoryError:
        typer.secho("Error: Path is a directory", fg=typer.colors.BRIGHT_RED)
        raise typer.Exit()
    except FileNotFoundError:
        typer.secho("Error: File not found", fg=typer.colors.BRIGHT_RED)
        raise typer.Exit()
    except SyntaxError as e:
        typer.secho(f"Syntax Error: {e}", fg=typer.colors.BRIGHT_RED)
        raise typer.Exit()



import os
from triton_copilot.init_services import save_openai_tokens, save_claude_tokens, is_initialized, get_model
from triton_copilot.run_services import wait_for_container_to_start, get_curl_command, run_docker_image, \
    echo_run_instructions
from triton_copilot.utils import rename_dir, cleanup
from triton_copilot.triton_services import create_triton_code, create_config_pbtxt, modify_modelpy, get_triton_inputs
from triton_copilot.build_services import build_image, get_function_refs
from triton_copilot.validate import validate_refs, check_dependencies, validate_file

import typer
from enum import Enum
from rich.progress import Progress, SpinnerColumn, TextColumn
import json
app = typer.Typer()


class ModelChoices(str, Enum):
    gpt4 = "gpt4"
    claude3 = "claude3"


@app.command(name="init")
def initialize():
    try:
        typer.secho("Welcome to Triton Copilot!", fg=typer.colors.CYAN)
        typer.secho("This tool helps you build and run Triton Inference Server images. Let's get started!", fg=typer.colors.CYAN)
        with Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                transient=True,
        ) as progress:
            progress.add_task("Checking dependencies", total=None)
            error = check_dependencies()
            if error:
                typer.secho(f"Please install the missing dependencies and try again. {error}", fg=typer.colors.BRIGHT_RED)
                # raise typer.Exit()
            typer.echo("Dependencies validated successfully\n")

        typer.secho("We support OpenAI's gpt-4-turbo-preview, Anthropic's claude-3-opus-20240229 models to generate "
                    "triton compatible code.", fg=typer.colors.MAGENTA)
        model = typer.prompt("Select the model you want to build Triton image for [gpt4, claude3]", type=ModelChoices)
        if model == "gpt4":
            openai_key = typer.prompt("Enter your OpenAI API key")
            openai_org = typer.prompt("Enter your OpenAI organization ID", default="None")
            if openai_org == "None":
                openai_org = None
            save_openai_tokens(openai_key, openai_org)
        elif model == "claude3":
            claude_api_key = typer.prompt("Enter your Claude API key")
            save_claude_tokens(claude_api_key)

        typer.echo("You are all set to build Triton Server image")
    except Exception as e:
        typer.secho(f"Something went wrong: {e}", fg=typer.colors.BRIGHT_RED)


@app.command(name="build")
def build_triton(
        file_path: str = typer.Argument(help="Path to the python file"),
        tag: str = typer.Option(help="Docker image tag", prompt="Enter the image tag"),
        triton_version: str = typer.Option("24.05-trtllm-python-py3", help="Triton version", prompt="Enter the Triton version"),
        load_ref: str = typer.Option(default="None", help="Reference to the load function"),
        infer_ref: str = typer.Option(default="None", help="Reference to the inference function"),
        unload_ref: str = typer.Option(default="None", help="Reference to the unload function"),
        sample_input_file: str = typer.Option("None", help="file path to sample input payload"),
        sample_output_file: str = typer.Option("None", help="file path to Sample output payload"),
        requirements_file: str = typer.Option("None", help="file path to requirements.txt"),
        max_batch_size: int = typer.Option(None, help="Max batch size"),
        preferred_batch_size: int = typer.Option(None, help="Preferred batch size"),
        no_prompt: bool = typer.Option(False, help="Skip prompts"),
):
    try:
        if not is_initialized():
            typer.secho("Please run 'triton-copilot init' to initialize the tool", fg=typer.colors.BRIGHT_RED)
            typer.Exit()
        validate_file(file_path)
        if not no_prompt:
            if load_ref == "None":
                load_ref, infer_ref, unload_ref = get_function_refs(file_path)
            if sample_input_file == "None":
                sample_input_file = typer.prompt("File path of sample input payload", default="None")
            if sample_output_file == "None":
                sample_output_file = typer.prompt("File path of sample output payload", default="None")
            if requirements_file == "None":
                directory = "." if os.path.dirname(file_path)=="" else os.path.dirname(file_path)
                if "requirements.txt" in os.listdir(directory):
                    requirements_file = typer.prompt("File path of requirements.txt", default="requirements.txt")
                else:
                    requirements_file = typer.prompt("File path of requirements.txt", default="None")
            if not max_batch_size:
                batching = typer.confirm("Do you want to enable batching?")
                max_batch_size = None
                preferred_batch_size = None
                if batching:
                    max_batch_size = typer.prompt("Enter the max batch size", default=2)
                    preferred_batch_size = typer.prompt("Enter the preferred batch size", default=2)
        validate_refs(load_ref, infer_ref, unload_ref, file_path)
        model = get_model()
        with Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                transient=True,
        ) as progress:
            progress.add_task("Generating Triton code\n", total=None)
            project_path = create_triton_code(file_path, load_ref, infer_ref, unload_ref)
            model_name = create_config_pbtxt(model, file_path, project_path, sample_input_file, sample_output_file, max_batch_size, preferred_batch_size)
            inputs, outputs = modify_modelpy(project_path, model)
            triton_inputs = get_triton_inputs(model, inputs)
            temp_path = project_path
            project_path = rename_dir(project_path, model_name)
            progress.add_task("Building image\n", total=None)
            build_image(project_path, tag, triton_version, requirements_file)
            cleanup(temp_path)
            cleanup(project_path)
        typer.echo(f"Image built successfully with tag: {tag}")
        echo_run_instructions(tag, triton_inputs, model_name)
    except Exception as e:
        typer.secho(f"Something went wrong: {e}", fg=typer.colors.BRIGHT_RED)
        cleanup(temp_path) if "temp_path" in locals() else None
        cleanup(project_path) if "project_path" in locals() else None
    except KeyboardInterrupt or SystemExit:
        typer.secho("Exiting...", fg=typer.colors.BRIGHT_RED)
        cleanup(temp_path) if "temp_path" in locals() else None
        cleanup(project_path) if "project_path" in locals() else None


@app.command(name="run")
def run_triton_image(
        tag: str = typer.Argument(help="Docker image tag"),
        volume: str = typer.Option("None", help="Volume name"),
        env: str = typer.Option("None", help="Environment variables as a json"),
):
    try:
        with Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                transient=True,
        ) as progress:
            if not is_initialized():
                typer.secho("Please run 'triton-copilot init' to initialize the tool", fg=typer.colors.BRIGHT_RED)
                typer.Exit()
            if volume == "None":
                volume = None
            if env == "None":
                env_vars = None
            else:
                env_vars = json.loads(env)
            task_id = progress.add_task("Running Triton Inference Server", total=None)
            port = run_docker_image(tag, volume, env_vars)
            progress.stop_task(task_id)
            task_id = progress.add_task("Waiting for Triton Inference Server to start", total=None)
            is_started = wait_for_container_to_start(port)
            progress.stop_task(task_id)
            if is_started:
                typer.echo("Triton Inference Server is running")
                get_curl_command(port)
            else:
                typer.secho("Triton Inference Server failed to start in 300s, please check docker logs",
                            fg=typer.colors.BRIGHT_RED)
    except Exception as e:
        typer.secho(f"Something went wrong: {e}", fg=typer.colors.BRIGHT_RED)


if __name__ == "__main__":
    app()



import os
import typer
from triton_copilot.constants import system_prompt_config_pbtxt
from triton_copilot.init_services import get_claude_keys, get_openai_keys
from triton_copilot.utils import copy_code_to_new_dir, claude_chat, openai_chat, get_model_name, save_config, cleanup
from triton_copilot.validate import validate_file
import pkg_resources
import ast
import uuid

TEMPLATE_DIR = pkg_resources.resource_filename("triton_copilot", "template/")


def get_user_prompt(file_path, sample_input, sample_output):
    with open(file_path, "r") as f:
        app_code = f.read()
    user_prompt_text = f"Provided Code Snippet: \n {app_code}"
    if sample_input:
        user_prompt_text += f"\nUser has provided a sample_input. The contents are: \n {sample_input}"
    elif sample_output:
        user_prompt_text += f"\nUser has provided a sample_output. The contents are: \n {sample_output}"

    user_prompt = [
        {
            "role": "user",
            "content": [
                {
                    "type": "text",
                    "text": user_prompt_text
                }
            ]
        }
    ]
    user_prompt.append(
        {
            "role": "assistant",
            "content": "The config.pbtxt is:"
        }
    )
    return user_prompt


def map_input_output_data_types(input_output):
    model_triton_data_type_maping = {
        "TYPE_BOOL": "BOOL",
        "TYPE_UINT8": "UINT8",
        "TYPE_UINT16": "UINT16",
        "TYPE_UINT32": "UINT32",
        "TYPE_UINT64": "UINT64",
        "TYPE_INT8": "INT8",
        "TYPE_INT16": "INT16",
        "TYPE_INT32": "INT32",
        "TYPE_INT64": "INT64",
        "TYPE_FP16": "FP16",
        "TYPE_FP32": "FP32",
        "TYPE_FP64": "FP64",
        "TYPE_STRING": "BYTES",
        "TYPE_BF16": "BF16",
    }
    for input in input_output["inputs"]:
        input["data_type"] = model_triton_data_type_maping[input["data_type"]]
    for output in input_output["outputs"]:
        output["data_type"] = model_triton_data_type_maping[output["data_type"]]

    return input_output["inputs"], input_output["outputs"]


def get_input_output(file_path, model):
    with open(file_path, "r") as f:
        data = f.read()

    if model == "gpt4":
        messages = [
            {
                "role": "system",
                "content": """
                    Extract the input and output as dictionaries from the provided text.
                    The output should be a dictionary with the keys "inputs" and "outputs".
                    Both input and output should be a list of dictionaries
                """
            },
            {
                "role": "user",
                "content": data
            }
        ]
        api_key, org_id = get_openai_keys()
        input_output = openai_chat(messages, api_key, org_id)
        input_output = ast.literal_eval(input_output)
    elif model == "claude3":
        system_message = """
            Extract the input and output as dictionaries from the provided text.
            The output should be a dictionary with the keys "inputs" and "outputs".
            Both input and output should be a list of dictionaries
        """
        messages = [
            {
                "role": "user",
                "content": data
            },
            {
                "role": "assistant",
                "content": "The extracted input and output is:"
            }
        ]
        api_key = get_claude_keys()
        input_output = claude_chat(system_message, messages, api_key)
        input_output = ast.literal_eval(input_output)

    return map_input_output_data_types(input_output)


def get_triton_inputs(model, inputs):
    try:
        if model == "gpt4":
            messages = [
                {
                    "role": "system",
                    "content": """
                        Generate sample triton inputs payload for given schema. 
                        Return a dictionary only with "inputs" as the key.
                        This object should be a valid python dictionary without any comments or extra text.
                        Use only doule quotes in the response, do not use single quotes at all
                    """
                },
                {
                    "role": "user",
                    "content": str(inputs)
                }
            ]
            api_key, org_id = get_openai_keys()
            triton_inputs = openai_chat(messages, api_key, org_id)
            triton_inputs = ast.literal_eval(triton_inputs)
        elif model == "claude3":
            system_message = """
                Generate sample triton inputs payload for given schema. 
                Return a dictionary only with "inputs" as the key.
                This object should be a valid python dictionary without any comments or extra text.
                Use only doule quotes in the response, do not use single quotes at all
            """
            messages = [
                {
                    "role": "user",
                    "content": str(inputs)
                },
                {
                    "role": "assistant",
                    "content": "The extracted input is:"
                }
            ]
            api_key = get_claude_keys()
            triton_inputs = claude_chat(system_message, messages, api_key)
            triton_inputs = ast.literal_eval(triton_inputs)

        return triton_inputs
    except ValueError as e:
        typer.secho(f"generated triton inputs are not in the correct format, try again: {e}", fg=typer.colors.BRIGHT_RED)
        raise typer.Exit()
    except SyntaxError as e:
        typer.secho(f"generated triton inputs are not in the correct format, try again: {e}", fg=typer.colors.BRIGHT_RED)
        raise typer.Exit()
    except Exception as e:
        typer.secho(f"Failed to get triton inputs: {e}", fg=typer.colors.BRIGHT_RED)
        raise typer.Exit()


def modify_modelpy(directory, model):
    try:
        inputs, outputs = get_input_output(f"{directory}/config.pbtxt", model)
        for each in inputs:
            each["datatype"] = each.pop("data_type")
        for each in outputs:
            each["datatype"] = each.pop("data_type")

        file = open(f"{directory}/1/model.py", "r")
        lines = file.readlines()
        file.close()
        for i in range(len(lines)):
            lines[i] = lines[i].replace('["##input_list##"]', str(inputs))
            lines[i] = lines[i].replace('["#output_list#"]', str(outputs))

        file = open(f"{directory}/1/model.py", "w")
        file.writelines(lines)

        return inputs, outputs
    except Exception as e:
        typer.secho(f"Failed at creating model.py: {e}", fg=typer.colors.BRIGHT_RED)
        raise typer.Exit()


def create_triton_code(file_path, load_ref, infer_ref, unload_ref):
    try:
        base_dir = os.path.dirname(file_path)
        if base_dir == "":
            base_dir = "."
        os.makedirs("./temp/1", exist_ok=True)
        temp_dir = "./temp"
        copy_code_to_new_dir(base_dir, temp_dir + "/1")
        if file_path.split("/")[-1] == "model.py":
            rand_id = uuid.uuid4().hex[:4]
            file_path = f"{temp_dir}/1/model_{rand_id}.py"
            os.rename(f"{temp_dir}/1/model.py", file_path)

        # check if there is a file called model.py in the directory
        if "model.py" in os.listdir(os.path.dirname(temp_dir + "/1")):
            typer.secho("Error: model.py file found in the directory, please do not use `model.py` as "
                        "module name as it is reserved for triton entrypoint ", fg=typer.colors.BRIGHT_RED)
            raise typer.Exit()

        copy_code_to_new_dir(TEMPLATE_DIR, temp_dir)
        cleanup(f"{temp_dir}/1/temp")
        file = open(f"{temp_dir}/1/model.py", "r")
        lines = file.readlines()
        for i in range(len(lines)):
            lines[i] = lines[i].replace("##model_class##", load_ref.split(".")[0])
            lines[i] = lines[i].replace("##load_func##", load_ref.split(".")[1])
            lines[i] = lines[i].replace("##infer_func##", infer_ref.split(".")[1])
            lines[i] = lines[i].replace("##file##", file_path.split("/")[-1].split(".")[0])
            if unload_ref is not None and unload_ref != "None":
                lines[i] = lines[i].replace("##unload_func##", unload_ref.split(".")[1])

        if unload_ref is None or unload_ref == "None":
            lines = [line for line in lines if "finalize" not in line and "unload" not in line]

        file = open(f"{temp_dir}/1/model.py", "w")
        file.writelines(lines)
        return temp_dir
    except Exception as e:
        typer.secho(f"Failed to create triton code: {e}", fg=typer.colors.BRIGHT_RED)
        raise typer.Exit()


def create_config_pbtxt(model, file_path, project_path, sample_input_file, sample_output_file, max_batch_size, preferred_batch_size):
    try:
        if sample_input_file == "None":
            sample_input = None
        else:
            validate_file(sample_input_file)
            with open(sample_input_file, "r") as f:
                sample_input = f.read()
        if sample_output_file == "None":
            sample_output = None
        else:
            validate_file(sample_output_file)
            with open(sample_output_file, "r") as f:
                sample_output = f.read()
        output = ""
        if model == "gpt4":
            api_key, org_id = get_openai_keys()
            messages = [
                {
                    "role": "system",
                    "content": system_prompt_config_pbtxt
                }
            ]
            user_prompt = get_user_prompt(file_path, sample_input, sample_output)
            messages.extend(user_prompt)
            output = openai_chat(messages, api_key, org_id)
        elif model == "claude3":
            api_key = get_claude_keys()
            new_dir, user_prompt = get_user_prompt(file_path, sample_input, sample_output)
            output = claude_chat(system_prompt_config_pbtxt, user_prompt, api_key)

        if max_batch_size is not None:
            max_queue_delay = 500000
            output = output + "\ndynamic_batching {\n"
            output = output + f"  preferred_batch_size: [ {preferred_batch_size} ]\n"
            output = output + f"  max_queue_delay_microseconds: {max_queue_delay}\n"
            output = output + "}\n"
            output = output + f"max_batch_size: {max_batch_size}\n"

        save_config(output, project_path)
        return get_model_name(output)
    except FileNotFoundError:
        typer.secho("Error: Sample input or output file not found", fg=typer.colors.BRIGHT_RED)
        raise typer.Exit()
    except Exception as e:
        typer.secho(f"Failed to create config.pbtxt: {e}", fg=typer.colors.BRIGHT_RED)
        raise typer.Exit()




import os
import socket
import subprocess

import anthropic
from openai import OpenAI
import pkg_resources
import shutil

TEMPLATE_DIR = pkg_resources.resource_filename("triton_copilot", "template/")


def copy_code_to_new_dir(src, dest):
    os.system(f"cp -r {src}/* {dest}")


def save_config(output, directory):
    with open(os.path.join(directory, "config.pbtxt"), "w") as f:
        f.write(output)


def cleanup_llm_response(response):
    lines = response.split("\n")
    i = 0
    while i < len(lines):
        if lines[i].startswith("```"):
            lines.pop(i)
        else:
            i += 1
    return "\n".join(lines)


def openai_chat(messages, api_key, org_id):
    client = OpenAI(organization=org_id, api_key=api_key)
    completion = client.chat.completions.create(
        model="gpt-4-turbo-preview",
        temperature=0.5,
        messages=messages,
    )
    return cleanup_llm_response(completion.choices[0].message.content)


def claude_chat(system_message, messages, api_key):
    client = anthropic.Anthropic(
        api_key=api_key,
    )
    message = client.messages.create(
        max_tokens=4096,
        model="claude-3-opus-20240229",
        system=system_message,
        messages=messages
    )
    return cleanup_llm_response(message.content[0].text)


def get_model_name(config_pbtxt):
    lines = config_pbtxt.split("\n")
    # 1st line in config.pbtxt looks like this: name: "gpt-4-turbo-preview"
    model_name = lines[0].replace("\"", "").split(":")[1].strip()
    return model_name


def rename_dir(directory, new_name):
    new_dir_name = "/".join(directory.split("/")[:-1] + [new_name])
    os.rename(directory, new_dir_name)
    return new_dir_name


def cleanup(directory):
    try:
        shutil.rmtree(directory)  # Delete the directory and its contents recursively
    except FileNotFoundError:
        pass


def get_free_ports():
    start = 8000
    end = 8999
    ports = []
    for port in range(start, end + 1):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.bind(("localhost", port))
                s.listen(1)
                ports.append(port)
            except OSError:
                continue
        if len(ports) == 3:
            return ports
    raise RuntimeError(f"No free ports found in the range {start}-{end}")


def is_container_running(image_name):
    output = subprocess.run(["docker", "ps"], stdout=subprocess.PIPE)
    return image_name in output.stdout.decode()



import time
import requests
import typer
import subprocess
import json
from triton_copilot.utils import get_free_ports, is_container_running

def run_docker_image(tag, volume, env_vars):
    try:
        ports = get_free_ports()
        command = f"docker run --rm -d --gpus all -p {ports[0]}:8000 -p {ports[1]}:8001 -p {ports[2]}:8002"
        if volume is not None:
            command = f"{command} -v {volume}"
        if env_vars:
            env_vars_str = " ".join(f"-e {key}={value}" for key, value in env_vars.items())
            command = f"{command} {env_vars_str}"
        command = f"{command} {tag}"
        completed_process = subprocess.run(
            command.split(" "), stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True
        )
        build_output_lines = completed_process.stderr.decode().split('\n')
        time.sleep(10)
        if is_container_running(tag):
            return ports[0]
        else:
            typer.secho("\nFailed to start Triton Inference Server, Build logs:\n", fg=typer.colors.BRIGHT_RED)
            for line in build_output_lines:
                typer.secho(line)
            raise RuntimeError()
    except subprocess.CalledProcessError as e:
        raise RuntimeError("command '{}' return with error (code {}): {}, error: {}".format(e.cmd, e.returncode, e.output, e.stderr))


def wait_for_container_to_start(port):
    url = f"http://localhost:{port}/v2/health/ready"
    wait_time = 300
    while wait_time > 0:
        try:
            response = requests.get(url)
            if response.status_code == 200:
                return True
            else:
                time.sleep(10)
                wait_time -= 10
        except Exception as e:
            time.sleep(10)
            wait_time -= 10

    return False


def get_curl_command(port):
    url = f"http://localhost:{port}/v2/repository/index"
    response = requests.post(url)
    if response.status_code == 200:
        data = response.json()
        if len(data) == 0:
            typer.secho("No models found", fg=typer.colors.BRIGHT_RED)
            typer.Exit()
        for model in data:
            if model.get("state") == "READY":
                model_name = model.get("name")
                model_version = model.get("version")
                typer.secho(f"Model name: {model_name}, Model version: {model_version} is ready, \n curl command to infer:")
                typer.secho(f"curl -X POST http://localhost:{port}/v2/models/{model_name}/versions/{model_version}/"
                            f"infer -H 'Content-Type: application/json' -d <payload>", fg=typer.colors.BRIGHT_GREEN)
                typer.secho("Replace <payload> with the actual payload!! \nsample payload was provided as part of the build step", fg=typer.colors.YELLOW)
            else:
                typer.secho(f"Model {model.get('name')} is not ready", fg=typer.colors.BRIGHT_YELLOW)
    else:
        typer.secho("Failed to get model details", fg=typer.colors.BRIGHT_RED)
        typer.Exit()


def echo_run_instructions(tag, triton_inputs, model_name):
    typer.secho(f"Command to run the Triton Inference Server:", fg=typer.colors.BRIGHT_GREEN)
    typer.echo(f"triton-copilot run {tag}")

    typer.secho(f"Curl Command to infer:", fg=typer.colors.BRIGHT_GREEN)
    typer.echo(f"curl -X POST http://localhost:8000/v2/models/{model_name}/versions/1/infer -H 'Content-Type: application/json' -d '{json.dumps(triton_inputs)}'")

